@@ -18,8 +18,10 @@
  * along with bcal.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include <stdio.h>
 #include <ctype.h>
+#include <errno.h>
+#include <stdio.h>
+#include <string.h>
 #include <unistd.h>
 #include <readline/history.h>
 #include <readline/readline.h>
@@ -40,6 +42,7 @@
 #define _ALIGNMENT_MASK 0xF
 
 typedef unsigned char bool;
+typedef unsigned char uchar;
 typedef unsigned int uint;
 typedef unsigned long ulong;
 typedef unsigned long long ull;
@@ -51,41 +54,48 @@ typedef __uint128_t maxuint_t;
 typedef __uint64_t maxuint_t;
 #endif
 
+/* CHS representation */
 typedef struct {
 	ulong c;
 	ulong h;
 	ulong s;
 } t_chs;
 
-static char *VERSION = "1.9";
+/* Settings */
+typedef struct {
+	uchar bcmode  : 1;
+	uchar minimal : 1;
+	uchar repl    : 1;
+	uchar rsvd    : 3; /* Reserved for future usage */
+	uchar loglvl  : 2;
+} settings;
+
+static char *VERSION = "2.1";
 static char *units[] = {"b", "kib", "mib", "gib", "tib", "kb", "mb", "gb", "tb"};
+static char *logarr[] = {"ERROR", "WARNING", "INFO", "DEBUG"};
 
 static char *FAILED = "1";
 static char *PASSED = "\0";
 static char *curexpr = NULL;
+static char prompt[8] = "bcal> ";
 
 static char uint_buf[UINT_BUF_LEN];
 static char float_buf[FLOAT_BUF_LEN];
 
-static int minimal;
-static int repl;
-
 static Data lastres = {"\0", 0};
-
-static int cur_loglevel = INFO;
-static char *logarr[] = {"ERROR", "WARNING", "INFO", "DEBUG"};
+static settings cfg = {0, 0, 0, 0, INFO};
 
 static void debug_log(const char *func, int level, const char *format, ...)
 {
 	va_list ap;
 
-	va_start(ap, format);
-
 	if (level < 0 || level > DEBUG)
 		return;
 
-	if (level <= cur_loglevel) {
-		if (cur_loglevel == DEBUG) {
+	va_start(ap, format);
+
+	if (level <= cfg.loglvl) {
+		if (cfg.loglvl == DEBUG) {
 			fprintf(stderr, "%s(), %s: ", func, logarr[level]);
 			vfprintf(stderr, format, ap);
 		} else {
@@ -156,19 +166,25 @@ static size_t bstrlcpy(char *dest, const char *src, size_t n)
 
 /*
  * Try to evaluate en expression using bc
+ * If argument is NULL, global curexpr is picked
  */
-static int try_bc()
+static int try_bc(char *expr)
 {
 	pid_t pid;
 	int pipe_pc[2], pipe_cp[2], ret;
+	size_t len, count = 0;
 
-	if (!curexpr)
-		return -1;
+	if (!expr) {
+		if (curexpr)
+			expr = curexpr;
+		else
+			return -1;
+	}
 
-	log(DEBUG, "expression: \"%s\"\n", curexpr);
+	log(DEBUG, "expression: \"%s\"\n", expr);
 
 	if (pipe(pipe_pc) == -1 || pipe(pipe_cp) == -1) {
-		printf("Could not pipe\n");
+		log(ERROR, "pipe()!\n");
 		exit(EXIT_FAILURE);
 	}
 
@@ -195,17 +211,72 @@ static int try_bc()
 	}
 
 	/* parent */
-	char buffer[128] = "";
-	ret = write(pipe_pc[1], "scale=5\n", 8);
-	ret = write(pipe_pc[1], curexpr, strlen(curexpr));
-	ret = write(pipe_pc[1], "\n", 1);
-	ret = read(pipe_cp[0], buffer, sizeof(buffer));
+	if (write(pipe_pc[1], "scale=10\n", 9) != 9) {
+		log(ERROR, "write(1)! [%s]\n", strerror(errno));
+		exit(-1);
+	}
+
+	if (write(pipe_pc[1], "last=", 5) != 5) {
+		log(ERROR, "write(2)! [%s]\n", strerror(errno));
+		exit(-1);
+	}
+
+	if (lastres.p[0]) {
+		ret = strlen(lastres.p);
+		if (write(pipe_pc[1], lastres.p, ret) != ret) {
+			log(ERROR, "write(3)! [%s]\n", strerror(errno));
+			exit(-1);
+		}
+	} else {
+		if (write(pipe_pc[1], "0", 1) != 1) {
+			log(ERROR, "write(4)! [%s]\n", strerror(errno));
+			exit(-1);
+		}
+	}
+
+	if (write(pipe_pc[1], "\n", 1) != 1) {
+		log(ERROR, "write(5)! [%s]\n", strerror(errno));
+		exit(-1);
+	}
+
+	ret = strlen(expr);
+	if (write(pipe_pc[1], expr, ret) != ret) {
+		log(ERROR, "write(6)! [%s]\n", strerror(errno));
+		exit(-1);
+	}
+
+	if (write(pipe_pc[1], "\n", 1) != 1) {
+		log(ERROR, "write(7)! [%s]\n", strerror(errno));
+		exit(-1);
+	}
+
+	static char buffer[128];
+
+	ret = read(pipe_cp[0], buffer, sizeof(buffer) - 1);
+	if (ret == -1) {
+		log(ERROR, "read()! [%s]\n", strerror(errno));
+		exit(-1);
+	}
+
+	buffer[ret] = '\0';
 
 	if (buffer[0] != '(') {
 		printf("%s", buffer);
-		bstrlcpy(lastres.p, buffer, NUM_LEN);
-		/* remove newline */
-		lastres.p[strlen(lastres.p) - 1] = '\0';
+		len = bstrlcpy(lastres.p, buffer, NUM_LEN);
+
+		/* remove newline appended at the end of result by bc */
+		(len >= 2) ? (len -= 2) : (len = 0);
+		lastres.p[len] = '\0';
+
+		/* Trim the decimal part, if any */
+		while (count < len) {
+			if (lastres.p[count] == '.') {
+				lastres.p[count] = '\0';
+				break;
+			}
+
+			++count;
+		}
 		lastres.unit = 0;
 		log(DEBUG, "result: %s %d\n", lastres.p, lastres.unit);
 		return 0;
@@ -362,7 +433,7 @@ static maxuint_t strtouquad(char *token, char **pch)
 
 	char *ptr;
 	maxuint_t val = 0, prevval = 0;
-	uint base = 10, multiplier, digit, bits_used = 0;
+	uint base = 10, multiplier = 0, digit, bits_used = 0;
 	uint max_bit_len = sizeof(maxuint_t) << 3;
 
 	if (token[0] == '0') {
@@ -442,10 +513,11 @@ static maxuint_t convertbyte(char *buf, int *ret)
 	if (*pch) {
 		*ret = -1;
 		return 0;
-	} else
-		*ret = 0;
+	}
 
-	if (minimal) {
+	*ret = 0;
+
+	if (cfg.minimal) {
 		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
@@ -492,12 +564,13 @@ static maxuint_t convertkib(char *buf, int *ret)
 	if (*pch) {
 		*ret = -1;
 		return 0;
-	} else
-		*ret = 0;
+	}
+
+	*ret = 0;
 
 	maxuint_t bytes = (maxuint_t)(kib * 1024);
 
-	if (minimal) {
+	if (cfg.minimal) {
 		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
@@ -539,12 +612,13 @@ static maxuint_t convertmib(char *buf, int *ret)
 	if (*pch) {
 		*ret = -1;
 		return 0;
-	} else
-		*ret = 0;
+	}
+
+	*ret = 0;
 
 	maxuint_t bytes = (maxuint_t)(mib * (1 << 20));
 
-	if (minimal) {
+	if (cfg.minimal) {
 		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
@@ -586,12 +660,13 @@ static maxuint_t convertgib(char *buf, int *ret)
 	if (*pch) {
 		*ret = -1;
 		return 0;
-	} else
-		*ret = 0;
+	}
+
+	*ret = 0;
 
 	maxuint_t bytes = (maxuint_t)(gib * (1 << 30));
 
-	if (minimal) {
+	if (cfg.minimal) {
 		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
@@ -633,12 +708,13 @@ static maxuint_t converttib(char *buf, int *ret)
 	if (*pch) {
 		*ret = -1;
 		return 0;
-	} else
-		*ret = 0;
+	}
+
+	*ret = 0;
 
 	maxuint_t bytes = (maxuint_t)(tib * ((maxuint_t)1 << 40));
 
-	if (minimal) {
+	if (cfg.minimal) {
 		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
@@ -680,12 +756,13 @@ static maxuint_t convertkb(char *buf, int *ret)
 	if (*pch) {
 		*ret = -1;
 		return 0;
-	} else
-		*ret = 0;
+	}
+
+	*ret = 0;
 
 	maxuint_t bytes = (maxuint_t)(kb * 1000);
 
-	if (minimal) {
+	if (cfg.minimal) {
 		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
@@ -727,12 +804,13 @@ static maxuint_t convertmb(char *buf, int *ret)
 	if (*pch) {
 		*ret = -1;
 		return 0;
-	} else
-		*ret = 0;
+	}
+
+	*ret = 0;
 
 	maxuint_t bytes = (maxuint_t)(mb * 1000000);
 
-	if (minimal) {
+	if (cfg.minimal) {
 		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
@@ -774,12 +852,13 @@ static maxuint_t convertgb(char *buf, int *ret)
 	if (*pch) {
 		*ret = -1;
 		return 0;
-	} else
-		*ret = 0;
+	}
+
+	*ret = 0;
 
 	maxuint_t bytes = (maxuint_t)(gb * 1000000000);
 
-	if (minimal) {
+	if (cfg.minimal) {
 		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
@@ -821,12 +900,13 @@ static maxuint_t converttb(char *buf, int *ret)
 	if (*pch) {
 		*ret = -1;
 		return 0;
-	} else
-		*ret = 0;
+	}
+
+	*ret = 0;
 
 	maxuint_t bytes = (__uint128_t)(tb * 1000000000000);
 
-	if (minimal) {
+	if (cfg.minimal) {
 		printf("%s B\n", getstr_u128(bytes, uint_buf));
 		return bytes;
 	}
@@ -1016,7 +1096,8 @@ static bool lba2chs(char *lba, t_chs *p_chs)
 
 static void show_basic_sizes()
 {
-	printf("char       : %lu\n"
+	printf("---------------\n Storage sizes\n---------------\n"
+		"char       : %lu\n"
 		"short      : %lu\n"
 		"int        : %lu\n"
 		"long       : %lu\n"
@@ -1040,48 +1121,41 @@ static void show_basic_sizes()
 		sizeof(long double));
 }
 
+static void prompt_help()
+{
+	printf("prompt keys:\n\
+ b          toggle bc mode\n\
+ r          show result from last operation\n\
+ s          show sizes of storage types\n\
+ ?          show prompt help\n\
+ q/double ↵ quit program\n");
+}
+
 static void usage()
 {
-	printf("usage: bcal [-c N] [-f FORMAT] [-s bytes] [-m] [-b] [-d] [-h]\n\
-            [expression] [N [unit]] \n\n\
+	printf("usage: bcal [-c N] [-f loc] [-s bytes] [expr]\n\
+            [N [unit]] [-b [expr]] [-m] [-d] [-h]\n\n\
 Storage expression calculator.\n\n\
 positional arguments:\n\
- expression  evaluate storage arithmetic expression\n\
-             +, -, *, / operators with decimal or hex operands\n\
-             unit can be multiplied or divided by +ve integers\n\
-             +, -, / work with two unit operands\n\
-             Examples:\n\
-               bcal \"(5kb+2mb)/3\"\n\
-               bcal \"5 tb / 12\"\n\
-               bcal \"2.5mb*3\"\n\
-               bcal \"(2giB * 2) / 2kib\"\n\
- N [unit]    capacity in B/KiB/MiB/GiB/TiB/kB/MB/GB/TB\n\
-             see https://wiki.ubuntu.com/UnitsPolicy\n\
-             default unit is B (byte), case is ignored\n\
-             N can be decimal or '0x' prefixed hex value\n\n\
+ expr       expression in decimal/hex operands\n\
+ N [unit]   capacity in B/KiB/MiB/GiB/TiB/kB/MB/GB/TB\n\
+            see https://wiki.ubuntu.com/UnitsPolicy\n\
+            default unit is B (byte), case is ignored\n\
+            N can be decimal or '0x' prefixed hex value\n\n\
 optional arguments:\n\
- -c N        show +ve integer N in binary, decimal and hex\n\
- -f FORMAT   convert CHS to LBA or LBA to CHS\n\
-             formats are hyphen-separated\n\
-             LBA format:\n\
-               starts with 'l':\n\
-               lLBA-MAX_HEAD-MAX_SECTOR\n\
-             CHS format:\n\
-               starts with 'c':\n\
-               cC-H-S-MAX_HEAD-MAX_SECTOR\n\
-             omitted values are considered 0\n\
-             FORMAT 'c-50--0x12-' denotes:\n\
-               C = 0, H = 50, S = 0, MH = 0x12, MS = 0\n\
-             FORMAT 'l50-0x12' denotes:\n\
-               LBA = 50, MH = 0x12, MS = 0\n\
-             default MAX_HEAD: 16, default MAX_SECTOR: 63\n\
- -s bytes    sector size [default 512]\n\
- -m          show minimal output (e.g. decimal bytes)\n\
- -b          list sizes of basic data types in bytes\n\
- -d          enable debug information and logs\n\
- -h          show this help and exit\n\n\
-Version %s\n\
-Copyright © 2016-2018 Arun Prakash Jana <engineerarun@gmail.com>\n\
+ -c N       show +ve integer N in binary, decimal, hex\n\
+ -f loc     convert CHS to LBA or LBA to CHS\n\
+            refer to the operational notes in man page\n\
+ -s bytes   sector size [default 512]\n\
+ -b [expr]  enter bc mode or evaluate expression in bc\n\
+ -m         show minimal output (e.g. decimal bytes)\n\
+ -d         enable debug information and logs\n\
+ -h         show this help\n\n");
+
+	prompt_help();
+
+	printf("\nVersion %s\n\
+Copyright © 2016 Arun Prakash Jana <engineerarun@gmail.com>\n\
 License: GPLv3\n\
 Webpage: https://github.com/jarun/bcal\n", VERSION);
 }
@@ -1103,8 +1177,8 @@ static maxuint_t unitconv(Data bunit, char *isunit, int *out)
 	/* Data is a C structure containing a string p and a char
 	 * indicating if the string is a unit or a plain number
 	 */
-	char *numstr = bunit.p, *punit;
-	int len, count;
+	char *numstr = bunit.p, *punit = NULL;
+	int  count;
 	maxfloat_t byte_metric = 0;
 
 	if (numstr == NULL || *numstr == '\0') {
@@ -1113,28 +1187,19 @@ static maxuint_t unitconv(Data bunit, char *isunit, int *out)
 		return 0;
 	}
 
+	log(DEBUG, "numstr: %s\n", numstr);
 	*out = 0;
-	len = strlen(numstr) - 1;
-	if (isdigit(numstr[len])) {
-		char *pc = NULL;
-		maxuint_t r;
 
-		/* ensure this is not the result of a previous operation */
-		if (*isunit != 1)
-			*isunit = 0;
-		r = strtoull(numstr, &pc, 0);
-		if (*numstr != '\0' && *pc == '\0')
-			return r;
+	/* ensure this is not the result of a previous operation */
+	if (*isunit != 1)
+		*isunit = 0;
 
-		log(ERROR, "invalid token: %s\n", pc);
-		*out = -1;
-		return 0;
-	}
+	byte_metric = strtold(numstr, &punit);
+	log(DEBUG, "byte_metric: %Lf\n", byte_metric);
+	if (*numstr != '\0' && *punit == '\0')
+		return (maxuint_t)byte_metric;
 
-	while (isalpha(numstr[len]))
-		--len;
-
-	punit = numstr + len + 1;
+	log(DEBUG, "punit: %s\n", punit);
 
 	count = ARRAY_SIZE(units);
 	while (--count >= 0)
@@ -1142,85 +1207,59 @@ static maxuint_t unitconv(Data bunit, char *isunit, int *out)
 			break;
 
 	if (count == -1) {
-		if (minimal)
+		if (cfg.minimal)
 			log(ERROR, "unknown unit\n");
 		else
-			try_bc();
+			try_bc(NULL);
 
 		*out = -1;
 		return 0;
 	}
 
 	*isunit = 1;
-	*punit = '\0';
 
 	switch (count) {
-	case 0: {
-		maxuint_t bytes;
-		bytes = (maxuint_t)strtold(numstr, &punit);
-		if (*punit)
-			goto error;
-		return bytes;
-		}
+	case 0:
+		return (maxuint_t)byte_metric;
 	case 1: /* Kibibyte */
-		byte_metric = strtold(numstr, &punit);
-		if (*punit)
-			goto error;
 		return (maxuint_t)(byte_metric * 1024);
 	case 2: /* Mebibyte */
-		byte_metric = strtold(numstr, &punit);
-		if (*punit)
-			goto error;
 		return (maxuint_t)(byte_metric * (1 << 20));
 	case 3: /* Gibibyte */
-		byte_metric = strtold(numstr, &punit);
-		if (*punit)
-			goto error;
 		return (maxuint_t)(byte_metric * (1 << 30));
 	case 4: /* Tebibyte */
-		byte_metric = strtold(numstr, &punit);
-		if (*punit)
-			goto error;
 		return (maxuint_t)(byte_metric * ((maxuint_t)1 << 40));
 	case 5: /* Kilobyte */
-		byte_metric = strtold(numstr, &punit);
-		if (*punit)
-			goto error;
 		return (maxuint_t)(byte_metric * 1000);
 	case 6: /* Megabyte */
-		byte_metric = strtold(numstr, &punit);
-		if (*punit)
-			goto error;
 		return (maxuint_t)(byte_metric * 1000000);
 	case 7: /* Gigabyte */
-		byte_metric = strtold(numstr, &punit);
-		if (*punit)
-			goto error;
 		return (maxuint_t)(byte_metric * 1000000000);
 	case 8: /* Terabyte */
-		byte_metric = strtold(numstr, &punit);
-		if (*punit)
-			goto error;
 		return (maxuint_t)(byte_metric * 1000000000000);
 	default:
 		log(ERROR, "unknown unit\n");
 		*out = -1;
 		return 0;
 	}
-
-error:
-	log(ERROR, "malformed input\n");
-	*out = -1;
-	return 0;
 }
 
-static int priority(char sign) /* Get the priority of operators */
+/* Get the priority of operators.
+ * Higher priority, higher value.
+ */
+static int priority(char sign) /* Get the priority of operators, higher priprity */
 {
 	switch (sign) {
-	case '-': return 1;
-	case '+': return 2;
-	case '*': return 3;
-	case '/': return 4;
+	case '|': return 1;
+	case '^': return 2;
+	case '&': return 3;
+	case '>':
+	case '<': return 4;
+	case '-':
+	case '+': return 5;
+	case '%':
+	case '/':
+	case '*': return 6;
 	}
 
 	return 0;
@@ -1231,9 +1270,15 @@ static int infix2postfix(char *exp, queue **resf, queue **resr)
 {
 	stack *op = NULL;  /* Operator Stack */
 	char *token = strtok(exp, " ");
-	Data tokenData = {"\0", 0};
+	static Data tokenData, ct;
 	int balanced = 0;
-	Data ct;
+	bool tokenize = TRUE;
+
+	tokenData.p[0] = '\0';
+	tokenData.unit = 0;
+
+	log(DEBUG, "exp: %s\n", exp);
+	log(DEBUG, "token: %s\n", token);
 
 	while (token) {
 		/* Copy argument to string part of the structure */
@@ -1244,6 +1289,12 @@ static int infix2postfix(char *exp, queue **resf, queue **resr)
 		case '-':
 		case '*':
 		case '/':
+		case '%':
+		case '>':
+		case '<':
+		case '&':
+		case '|':
+		case '^':
 			if (token[1] != '\0') {
 				log(ERROR, "invalid token terminator\n");
 				emptystack(&op);
@@ -1285,11 +1336,30 @@ static int infix2postfix(char *exp, queue **resf, queue **resr)
 			enqueue(resf, resr, lastres);
 			break;
 		default:
+			/*
+			 * Check if unit is specified
+			 * This also guards against a case of 0xn b
+			 */
+			token = strtok(NULL, " ");
+			if (token && token[0] == 'b' && token[1] == '\0') {
+				tokenData.unit = 1;
+				log(DEBUG, "unit found\n");
+			} else
+				tokenize = FALSE; /* We already toknized here */
+
 			/* Enqueue operands */
+			log(DEBUG, "tokenData: %s %d\n", tokenData.p, tokenData.unit);
 			enqueue(resf, resr, tokenData);
+			if (tokenize)
+				tokenData.unit = 0;
 		}
 
-		token = strtok(NULL, " ");
+		if (tokenize)
+			token = strtok(NULL, " ");
+		else
+			tokenize = TRUE;
+
+		log(DEBUG, "token: %s\n", token);
 	}
 
 	while (!isempty(op)) {
@@ -1318,7 +1388,7 @@ static int validate_div(maxuint_t dividend, maxuint_t divisor, maxuint_t quotien
 	if (divisor * quotient < dividend) {
 		log(WARNING, "result truncated\n");
 
-		if (cur_loglevel == DEBUG) {
+		if (cfg.loglvl == DEBUG) {
 			printhex_u128(dividend);
 			printf(" (dividend)\n");
 			printhex_u128(divisor);
@@ -1378,19 +1448,49 @@ static maxuint_t eval(queue **front, queue **rear, int *out)
 			raw_c.unit = 0;
 
 			switch (arg.p[0]) {
+			case '>':
+			case '<':
+				if (raw_b.unit) {
+					log(ERROR, "unit mismatch in %c%c\n", arg.p[0], arg.p[0]);
+					goto error;
+				}
+
+				if (arg.p[0] == '>')
+					c = a >> b;
+				else
+					c = a << b;
+				raw_c.unit = raw_a.unit;
+				break;
 			case '+':
-				/* Check if both are units */
+			case '&':
+			case '|':
+			case '^':
+				/* Check if the units match */
 				if (raw_a.unit == raw_b.unit) {
-					c = a + b;
+					switch (arg.p[0]) {
+					case '+':
+						c = a + b;
+						break;
+					case '&':
+						c = a & b;
+						break;
+					case '|':
+						c = a | b;
+						break;
+					case '^':
+						c = a ^ b;
+						break;
+					}
+
 					if (raw_a.unit)
 						raw_c.unit = 1;
 					break;
 				}
 
-				log(ERROR, "unit mismatch in +\n");
+				log(ERROR, "unit mismatch in %c\n", arg.p[0]);
 				goto error;
 			case '-':
-				/* Check if both are units */
+				/* Check if the units match */
 				if (raw_a.unit == raw_b.unit) {
 					if (b > a) {
 						log(ERROR, "negative result\n");
@@ -1424,11 +1524,12 @@ static maxuint_t eval(queue **front, queue **rear, int *out)
 
 				if (raw_a.unit && raw_b.unit) {
 					c = a / b;
-					raw_c.unit = 0;
 
 					validate_div(a, b, c);
 					break;
-				} else if (!raw_b.unit) {
+				}
+
+				if (!raw_b.unit) {
 					c = a / b;
 					if (raw_a.unit)
 						raw_c.unit = 1;
@@ -1439,15 +1540,30 @@ static maxuint_t eval(queue **front, queue **rear, int *out)
 
 				log(ERROR, "unit mismatch in /\n");
 				goto error;
+			case '%':
+				if (b == 0) {
+					log(ERROR, "division by 0\n");
+					goto error;
+				}
+
+				if (!(raw_a.unit || raw_b.unit)) {
+					c = a % b;
+					break;
+				}
+
+				log(ERROR, "unit mismatch in modulo\n");
+				goto error;
 			}
 
 			/* Convert to string */
 			bstrlcpy(raw_c.p, getstr_u128(c, uint_buf), NUM_LEN);
+			log(DEBUG, "c: %s unit: %d\n", raw_c.p, raw_c.unit);
+
 			/* Push to stack */
 			push(&est, raw_c);
 
 		} else {
-			log(DEBUG, "in else (%s)\n", arg.p);
+			log(DEBUG, "pushing (%s %d)\n", arg.p, arg.unit);
 			push(&est, arg);
 		}
 	}
@@ -1475,10 +1591,21 @@ error:
 
 static int issign(char c)
 {
-	if (c == '+' || c == '-' || c == '*' || c == '/')
+	switch (c) {
+	case '+':
+	case '-':
+	case '*':
+	case '/':
+	case '%':
+	case '>':
+	case '<':
+	case '&':
+	case '|':
+	case '^':
 		return 1;
-
-	return 0;
+	default:
+		return 0;
+	}
 }
 
 /* Check if a char is operator or not */
@@ -1489,6 +1616,12 @@ static int isoperator(char c)
 	case '-':
 	case '*':
 	case '/':
+	case '%':
+	case '>':
+	case '<':
+	case '&':
+	case '|':
+	case '^':
 	case '(':
 	case ')': return 1;
 	default: return 0;
@@ -1544,7 +1677,8 @@ static void removeinnerspaces(char *s)
 	char *p = s;
 
 	while (*s != '\0') {
-		if (!isspace(*s)) {
+		/* We should not combine 0xn b*/
+		if (!isspace(*s) || (*(s + 1) == 'b')) {
 			*p = *s;
 			++p;
 		}
@@ -1603,6 +1737,25 @@ static char *fixexpr(char *exp, int *unitless)
 		     isoperator(exp[i + 1]))) ||
 		    (isalpha(exp[i]) && isoperator(exp[i + 1])) ||
 		    (isoperator(exp[i]) && (exp[i + 1] == 'r'))) {
+			if (exp[i] == '<' || exp[i] == '>') { /* handle shift operators << and >> */
+				if (prev != exp[i] && exp[i] != exp[i + 1]) {
+					log(ERROR, "invalid operator %c\n", exp[i]);
+					*unitless = 0;
+					free(parsed);
+					return NULL;
+				}
+
+				if (prev == exp[i + 1]) { /* handle <<< or >>> */
+					log(ERROR, "invalid sequence %c%c%c\n", prev, exp[i], exp[i + 1]);
+					*unitless = 0;
+					free(parsed);
+					return NULL;
+				}
+
+				if (exp[i] == exp[i + 1])
+					goto loop_end;
+			}
+
 			parsed[j] = exp[i];
 			++j;
 			parsed[j] = ' ';
@@ -1613,6 +1766,7 @@ static char *fixexpr(char *exp, int *unitless)
 			++j;
 		}
 
+loop_end:
 		prev = exp[i];
 		++i;
 	}
@@ -1689,7 +1843,7 @@ static int convertunit(char *value, char *unit, ulong sectorsz)
 
 	log(DEBUG, "%s %s\n", value, units[count]);
 
-	if (!minimal && unit)
+	if (!cfg.minimal && unit)
 		printf("\033[1mUNIT CONVERSION\033[0m\n");
 
 	switch (count) {
@@ -1726,10 +1880,10 @@ static int convertunit(char *value, char *unit, ulong sectorsz)
 	}
 
 	if (ret == -1) {
-		if (minimal || unit) /* For running python test cases */
+		if (cfg.minimal || unit) /* For running python test cases */
 			log(ERROR, "malformed input\n");
 		else
-			return try_bc();
+			return try_bc(NULL);
 
 		return -1;
 	}
@@ -1738,7 +1892,7 @@ static int convertunit(char *value, char *unit, ulong sectorsz)
 	lastres.unit = 1;
 	log(DEBUG, "result: %s %d\n", lastres.p, lastres.unit);
 
-	if (minimal)
+	if (cfg.minimal)
 		return 0;
 
 	printf("\nADDRESS\n (d) %s\n (h) ",
@@ -1769,6 +1923,8 @@ static int evaluate(char *exp, ulong sectorsz)
 	char *expr = fixexpr(exp, &ret);  /* Make parsing compatible */
 	char *ptr;
 
+	log(DEBUG, "expr: %s\n", expr);
+
 	if (expr == NULL) {
 		if (ret)
 			return convertunit(exp, NULL, sectorsz);
@@ -1784,7 +1940,8 @@ static int evaluate(char *exp, ulong sectorsz)
 	bytes = eval(&front, &rear, &ret);  /* Evaluate Expression */
 	if (ret == -1)
 		return -1;
-	else if (ret == 1) {
+
+	if (ret == 1) {
 		ptr = getstr_u128(bytes, uint_buf);
 		printf("%s\n", ptr);
 		bstrlcpy(lastres.p, getstr_u128(bytes, uint_buf), UINT_BUF_LEN);
@@ -1793,7 +1950,7 @@ static int evaluate(char *exp, ulong sectorsz)
 		return 0;
 	}
 
-	if (!(minimal || repl))
+	if (!(cfg.minimal || cfg.repl))
 		printf("\033[1mRESULT\033[0m\n");
 
 	convertbyte(getstr_u128(bytes, uint_buf), &ret);
@@ -1807,7 +1964,7 @@ static int evaluate(char *exp, ulong sectorsz)
 	lastres.unit = 1;
 	log(DEBUG, "result2: %s %d\n", lastres.p, lastres.unit);
 
-	if (minimal)
+	if (cfg.minimal)
 		return 0;
 
 	printf("\nADDRESS\n (d) %s\n (h) ", ptr);
@@ -1817,6 +1974,41 @@ static int evaluate(char *exp, ulong sectorsz)
 	return 0;
 }
 
+int convertbase(char *arg)
+{
+	char *pch;
+
+	strstrip(arg);
+
+	if (*arg == '\0') {
+		log(ERROR, "empty input\n");
+		return -1;
+	}
+
+	if (*arg == '-') {
+		log(ERROR, "N must be >= 0\n");
+		return -1;
+	}
+
+	if (cfg.repl && arg[0] == 'r' && arg[1] == '\0')
+		arg = lastres.p;
+
+	maxuint_t val = strtouquad(arg, &pch);
+	if (*pch) {
+		log(ERROR, "invalid input\n");
+		return -1;
+	}
+
+	printf(" (b) ");
+	binprint(val);
+	printf("\n (d) %s\n (h) ",
+		getstr_u128(val, uint_buf));
+	printhex_u128(val);
+	printf("\n");
+
+	return 0;
+}
+
 int main(int argc, char **argv)
 {
 	int opt = 0, operation = 0;
@@ -1829,27 +2021,9 @@ int main(int argc, char **argv)
 		switch (opt) {
 		case 'c':
 		{
-			char *pch;
 			operation = 1;
-
-			if (*optarg == '-') {
-				log(ERROR, "N must be >= 0\n");
-				return -1;
-			}
-
-			maxuint_t val = strtouquad(optarg, &pch);
-			if (*pch) {
-				log(ERROR, "invalid input\n");
-				return -1;
-			}
-
-			printf("\033[1mBASE CONVERSION\033[0m\n");
-			printf(" (b) ");
-			binprint(val);
-			printf("\n (d) %s\n (h) ",
-				getstr_u128(val, uint_buf));
-			printhex_u128(val);
-			printf("\n\n");
+			convertbase(optarg);
+			printf("\n");
 			break;
 		}
 		case 'f':
@@ -1877,7 +2051,7 @@ int main(int argc, char **argv)
 				log(ERROR, "invalid input\n");
 			break;
 		case 'm':
-			minimal = 1;
+			cfg.minimal = 1;
 			break;
 		case 's':
 			if (*optarg == '-') {
@@ -1887,10 +2061,11 @@ int main(int argc, char **argv)
 			sectorsz = strtoul_b(optarg);
 			break;
 		case 'b':
-			show_basic_sizes();
+			cfg.bcmode = 1;
+			strncpy(prompt, "bc> ", 5);
 			break;
 		case 'd':
-			cur_loglevel = DEBUG;
+			cfg.loglvl = DEBUG;
 			log(DEBUG, "bcal v%s\n", VERSION);
 			log(DEBUG, "maxuint_t is %lu bytes\n", sizeof(maxuint_t));
 
@@ -1899,6 +2074,7 @@ int main(int argc, char **argv)
 			usage();
 			return 0;
 		default:
+			log(ERROR, "invalid option \'%c\'\n\n", (char)optopt);
 			usage();
 			return -1;
 		}
@@ -1907,34 +2083,46 @@ int main(int argc, char **argv)
 	log(DEBUG, "argc %d, optind %d\n", argc, optind);
 
 	if (!operation && (argc == optind)) {
-		char *tmp = NULL;
-		repl = 1;
+		char *ptr = NULL, *tmp = NULL;
+		cfg.repl = 1;
 		int enters = 0;
 
+		read_history(NULL);
+
 		printf("q/double Enter -> quit, ? -> help\n");
-		while ((tmp = readline("bcal> ")) != NULL) {
+		while ((tmp = readline(prompt)) != NULL) {
+			/* Quit on double Enter */
 			if (tmp[0] == '\0') {
-				if (enters == 1)
+				if (enters == 1) {
+					free(tmp);
 					break;
+				}
 
 				++enters;
+				free(tmp);
 				continue;
 			}
 
 			enters = 0;
 
+			/* Save the original pointer from readline() */
+			ptr = tmp;
+
 			strstrip(tmp);
 			if (tmp[0] == '\0') {
-				free(tmp);
+				free(ptr);
 				continue;
 			}
 
-			curexpr = tmp;
+			log(DEBUG, "ptr: [%s]\n", ptr);
+			log(DEBUG, "tmp: [%s]\n", ptr);
+
 			add_history(tmp);
 
-			if (tmp[1] == '\0') {
-				/* Show the last stored result */
-				if (tmp[0] == 'r') {
+			if ((strlen(tmp) == 1) && tmp[1] == '\0') {
+				switch (tmp[0]) {
+				case 'r':
+					/* Show the last stored result */
 					if (lastres.p[0] == '\0')
 						printf("no result stored\n");
 					else {
@@ -1944,38 +2132,73 @@ int main(int argc, char **argv)
 						printf("\n");
 					}
 
-					free(tmp);
+					free(ptr);
 					continue;
-				}
+				case 'b':
+					cfg.bcmode ^= 1;
+					if (cfg.bcmode) {
+						printf("bc vars: scale = 10, ibase = 10, last = r\n");
+						strncpy(prompt, "bc> ", 5);
+					} else
+						strncpy(prompt, "bcal> ", 7);
 
-				if (tmp[0] == '?') {
-					free(tmp);
-					usage();
+					free(ptr);
 					continue;
-				}
+				case '?':
+					prompt_help();
 
-				if (tmp[0] == 'q') {
-					free(tmp);
-					break;
+					free(ptr);
+					continue;
+				case 'q':
+					free(ptr);
+					write_history(NULL);
+					return 0;
+				case 's':
+					show_basic_sizes();
+
+					free(ptr);
+					continue;
+				default:
+					printf("invalid input\n");
+					free(ptr);
+					continue;
 				}
 			}
 
+			if (tmp[0] == 'c') {
+				convertbase(tmp + 1);
+				free(ptr);
+				continue;
+			}
+
+			if (cfg.bcmode) {
+				try_bc(tmp);
+				free(ptr);
+				continue;
+			}
+
+			curexpr = tmp;
+
 			/* Evaluate the expression */
 			evaluate(tmp, sectorsz);
 
-			free(tmp);
+			free(ptr);
 		}
 
+		write_history(NULL);
 		return 0;
 	}
 
-	/* Conversion */
+	/* Unit conversion */
 	if (argc - optind == 2)
 		if (convertunit(argv[optind], argv[optind + 1], sectorsz) == -1)
 			return -1;
 
 	/*Arithmetic operation*/
 	if (argc - optind == 1) {
+		if (cfg.bcmode)
+			return try_bc(argv[optind]);
+
 		curexpr = argv[optind];
 		return evaluate(argv[optind], sectorsz);
 	}
